{% extends "page.html" %}

{% block main %}
<div id="react-admin-hook">
  <script id="jupyterhub-admin-config">
    window.api_page_limit = parseInt("{{ api_page_limit|safe }}")
    window.base_url = "{{ base_url|safe }}"
  </script>
  <script src={{ static_url("js/admin-react.js") }}></script>
</div>
{% endblock %}

{% block footer %}
<div class="container-fluid navbar-default small version_footer">
  <div class="navbar-text">
    JupyterHub {{ server_version }}
  </div>
</div>
{% endblock %}
